/* tcp port forward utility

   Copyright (c) 2003 - 2005 Heikki Orsila <heikki.orsila@iki.fi>

   See COPYING file for terms (MIT license).
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include <errno.h>
#include <fcntl.h>
#include <signal.h>
#include <sys/poll.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <unistd.h>
#include <sys/time.h>
#include <sys/wait.h>
#include <netinet/in.h>

#include "net.h"
#include "pipe.h"
#include "accounting.h"
#include "version.h"
#include "support.h"
#include "prefix.h"


static struct net_host_id *allowed_hosts;
static int n_allowed_hosts;
static char *accounting_file;
static int connectionwait = -1;


static const char *syntax =
"SYNTAX:\n"
"\n"
"    tcp-pf -s port1 [-d host:port2] [-l multiplexfile -p port2] [--lazy]\n"
"    [-a allowed_host] [-r rate] [--insert file]\n"
"\n"
"OPTIONS:\n"
"\n"
" -a/--allow host   Allow incoming connections from host. This option can be\n"
"                   used many times to select many hosts.\n"
" -d host:port      Forward incoming connections to host:port\n"
" -h/--help         Print help\n"
" --insert file     Insert a prefix file to the beginning of a destination\n"
"                   data connection. This can be used to add a constant header\n"
"                   to a data connection.\n"
" --lazy            Let tcp connection hang until both ends have shutdown.\n"
" -l/--list file    Choose destination host from the given host file.\n"
"                   The host with least amount of transfered bytes is\n"
"                   selected. (network load balancing)\n"
" -p/--dport port   Set destination port (use with -l)\n"
" -r/--rate rate    Maximum number of bytes per second going through one\n"
"                   connection\n"
" -s port           Set source port to accept connections\n"
" -w x              Wait x seconds before connecting incoming connection to the\n"
"                   destination address. This could be used to slow port\n"
"                   scanners.\n";


static int get_host_port(char **address, char **port, char *in)
{
	char *sep;
	char *nstr;

	nstr = strdup(in);
	if (nstr == NULL)
		return 0;

	sep = strrchr(nstr, (int) ':');
	if (sep == NULL)
		return 0;

	*sep = 0;

	*address = strdup(nstr);
	*port = strdup(sep + 1);

	free(nstr);

	if (!(*address) || !(*port)) {
		free(*address);
		free(*port);
		return 0;
	}

	return 1;
}


static void handle_client(int src_fd, char *dst_host, char *dst_port)
{
	int dst_fd;
	struct pipe p1;
	struct pipe p2;
	char buf[1024];

	xsleep(connectionwait); /* does not sleep if < 0 */

	if (accounting_file) {
		long long bytes;
		bytes = get_account(buf, sizeof(buf), accounting_file);
		if (bytes < 0)
			die("no hosts found from host file %s\n", accounting_file);

		fprintf(stderr,	"choosing host %s with %.1f mebibytes (%lld bytes)\n",buf, ((double) bytes) / (1024.0 * 1024.0), bytes);

		dst_host = buf;
	}

	dst_fd = net_open(dst_host, dst_port, "tcp", 0);
	if (dst_fd < 0) {
		fprintf(stderr, "couldn't connect to %s:%s\n", dst_host, dst_port);
		return;
	}

	memset(&p1, 0, sizeof(struct pipe));
	memset(&p2, 0, sizeof(struct pipe));

	p1.src_fd = src_fd;
	p1.dst_fd = dst_fd;
	p2.src_fd = dst_fd;
	p2.dst_fd = src_fd;

	if (handle_prefix_file(dst_fd)) {
		/* here is the main loop copying data from one socket to another */
		pmr_loop(&p1, &p2, dst_host, accounting_file, NULL, NULL);
	}

	if (accounting_file != NULL) {
		long long bytes;
		long long incr = p1.accbytes + p2.accbytes;
		bytes = account_increment(dst_host, accounting_file, incr);
		if (bytes < 0)
			die("%s not found from host file. can not increment!\n", dst_host);

		fprintf(stderr,	"host %s incremented by %.1f mebibytes to %.1f mebibytes (%lld bytes)\n", dst_host, ((double) incr) / (1024.0 * 1024.0), ((double) bytes) / (1024.0 * 1024.0), bytes);
		p1.accbytes = p2.accbytes = 0;
	}

	close(p1.src_fd);
	close(p1.dst_fd);
	close(p2.src_fd);
	close(p2.dst_fd);
}


static int accept_loop(char *src_port, char *dst_host, char *dst_port)
{
	int server_fd;
	int src_fd;
	pid_t child;

	char sockbuf[256];
	struct sockaddr_in *sockname = (struct sockaddr_in *) sockbuf;
	socklen_t socklen;
	int i;
	struct net_host_id id;
	char printed_id[256];
	int got_info;

	server_fd = net_bind(0, 0, src_port, "tcp", 1 /* REUSE */);
	if (server_fd < 0)
		die("can't bind server port\n");

	if (listen(server_fd, 5))
		dieerror("can't listen on server port");

	while (1) {
		socklen = sizeof(sockbuf);
		src_fd = accept(server_fd, (struct sockaddr *) sockname, &socklen);
		if (src_fd < 0) {
			if (errno == EAGAIN || errno == EINTR)
				continue;

			perror("client accept failed");
			goto err_out;
		}

		got_info = net_get_id(&id, (struct sockaddr *) sockname);
		if (!got_info) {
			fprintf(stderr, "client connected. can't get info on socket. rejecting.\n");
			if (n_allowed_hosts > 0)
				goto reject;
		}

		if (got_info) {
			printed_id[0] = 0;
			net_print_id(printed_id, &id, sizeof(printed_id));

			if (n_allowed_hosts > 0) {
				for (i = 0; i < n_allowed_hosts; i++) {
					if (net_are_same_ips
					    (&id, allowed_hosts + i))
						goto accept;
				}
				fprintf(stderr,
					"client connected from %s => reject\n",
					printed_id);
				goto reject;
			}
		      accept:
			fprintf(stderr,
				"client connected from %s => accept\n",
				printed_id);
		}

		child = fork();
		if (!child) {
			close(server_fd);
			child = fork();
			if (!child) {
				handle_client(src_fd, dst_host, dst_port);
				printf("client handled\n");
			} else if (child == -1) {
				perror("children can not reproduce");
			}
			abort();

		} else if (child == -1) {
			perror("can not have children. client will die.");

		} else {
			wait(0);
		}

	      reject:

		close(src_fd);
	}

	return 0;

      err_out:
	close(server_fd);
	return 1;
}


static void print_help(void)
{
	printf("\n");
	printf("tcp-pf %s by %s\n", TOOL_VERSION, AUTHOR);
	printf("Project web page: %s\n", PROJECT_WEB);
	printf("\n");
	printf("%s", syntax);
}


int main(int argc, char **argv)
{
	int i;
	char *src_port = NULL;
	char *dst_host = NULL;
	char *dst_port = NULL;
	char *endptr;
	struct net_host_id id;
	struct net_host_id *new_allowed;

	if (argc == 1) {
		print_help();
		exit(1);
	}

	for (i = 1; i < argc;) {

		if (!strcmp("-s", argv[i])) {
			REQUIRE_ARG("give source port.\n");
			src_port = strdup(argv[i + 1]);
			if (src_port == NULL)
				die("not enough memory for src_port\n");

			i += 2;

		} else if (!strcmp("-d", argv[i])) {
			REQUIRE_ARG("give destination host:port\n");

			if (!get_host_port(&dst_host, &dst_port, argv[i + 1]))
				die("Invalid host:port (%s)\n", argv[i + 1]);

			i += 2;

		} else if (!strcmp("--list", argv[i])
			   || !strcmp("-l", argv[i])) {
			REQUIRE_ARG("give accounting file\n");
			accounting_file = strdup(argv[i + 1]);
			if (accounting_file == NULL)
				die("Not enough memory for accounting filename\n");

			i += 2;

		} else if (!strcmp("--dport", argv[i])
			   || !strcmp("-p", argv[i])) {
			REQUIRE_ARG("give dst port.\n");
			dst_port = strdup(argv[i + 1]);
			if (dst_port == NULL)
				die("not enough memory for dst port\n");
			i += 2;

		} else if (!strcmp("--lazy", argv[i])) {
			kill_on_eof = 0;
			i++;

		} else if (!strcmp("--allow", argv[i])
			   || !strcmp("-a", argv[i])) {
			REQUIRE_ARG("give allowed hostname.\n");

			if (!net_info(&id, argv[i + 1]))
				die("nothing known about %s\n", argv[i + 1]);

			new_allowed = realloc(allowed_hosts, (n_allowed_hosts + 1) * sizeof(struct net_host_id));
			if (!new_allowed)
				die("no memory for allowed hosts\n");

			allowed_hosts = new_allowed;
			allowed_hosts[n_allowed_hosts] = id;
			n_allowed_hosts++;
			i += 2;

		} else if (!strcmp("-r", argv[i])
			   || !strcmp("--rate", argv[i])) {
			REQUIRE_ARG("expecting a rate limit value (bytes per second) with -r\n");
			pmr_max_rate = atoi(argv[i + 1]);
			if (pmr_max_rate <= 0)
				die("illegal bytes per second value (%d)\n", pmr_max_rate);

			i += 2;

		} else if (!strcmp("-h", argv[i])
			   || !strcmp("--help", argv[i])) {
			print_help();
			return 0;

		} else if (strcmp("--insert", argv[i]) == 0) {
			REQUIRE_ARG("give prefix file name.\n");
			prefix_file = strdup(argv[i + 1]);
			if (prefix_file == NULL)
				die("Not enough memory for prefix file name.\n");

			i += 2;

		} else if (strcmp("-w", argv[i]) == 0) {
			REQUIRE_ARG("Give connection delay\n");

			connectionwait = strtol(argv[i + 1], &endptr, 10);
			if (connectionwait <= 0 || *endptr != 0)
				die("Invalid connection delay: %s\n", argv[i + 1]);

			i += 2;

		} else {
			die("illegal arg: %s\n", argv[i]);
		}
	}

	if (src_port == NULL)
		die("you must provide source port with -s port!\n");

	if (accounting_file == NULL) {
		if (dst_host == NULL || dst_port == NULL) {
			die("you must provide destination host:port with -d host:port, or alternatively\n"
			    "you can provide destination list and destination port with -l listfile -p port\n");
		}
	} else if (dst_port == NULL) {
		die("you must provide destination port with -p port\n");
	}

	return accept_loop(src_port, dst_host, dst_port);
}
