/* support tool for tcppipe to lookup a hostname with least amount of bytes
   from a hostfile.

   Copyright (c) 2003 - 2005 Heikki Orsila <heikki.orsila@iki.fi>

   See COPYING file for terms (MIT license).
*/

#include <stdlib.h>
#include <stdio.h>
#include <limits.h>
#include <string.h>

#include "accounting.h"
#include "version.h"

static void print_help(void)
{
	fprintf(stderr, "tcppipe-least %s by %s\n", TOOL_VERSION, AUTHOR);
	fprintf(stderr, "Project web page: %s\n", PROJECT_WEB);
	fprintf(stderr, "\n");
	fprintf(stderr,
		"USAGE: tcppipe-least [-c] [-i host bytes] [hostfile]\n");
	fprintf(stderr, "\n");
	fprintf(stderr, "OPTIONS\n");
	fprintf(stderr, "\n");
	fprintf(stderr,
		" -c               Print byte amount from the host file\n");
	fprintf(stderr,
		" -i host bytes    Adds a specified byte amount for the host.\n");
	fprintf(stderr, "\n");
	fprintf(stderr,
		"Minimum host is the host with least amount of bytes in the host file.\n");
	fprintf(stderr,
		"The host file can be given as the last argument, or otherwise\n");
	fprintf(stderr, "$(HOME)/.stt/default is used.\n");
	fprintf(stderr, "\n");
	fprintf(stderr,
		"Example 1: Print the minimum host from /tmp/hostfile:\n");
	fprintf(stderr, "\ttcppipe-least /tmp/hostfile\n");
	fprintf(stderr, "\n");
	fprintf(stderr,
		"Example 2: Increment host foo by 2^30 bytes (default host file)\n");
	fprintf(stderr, "\ttcppipe-least -i foo 1073741824\n");
	fprintf(stderr, "\n");
	fprintf(stderr,
		"Example 3: Print the minimum host with its byte amount (default host file)\n");
	fprintf(stderr, "\ttcppipe-least -c\n");
	fprintf(stderr, "\n");
}


int main(int argc, char **argv)
{
	char host[256];
	char file[PATH_MAX];
	int i;
	int increment_mode = 0;
	int64_t increment = 0;
	char *endptr;
	int print_bytes = 0;

	for (i = 1; i < argc;) {
		if (strcmp(argv[i], "-c") == 0) {
			print_bytes = 1;
			i++;
		} else if (strcmp(argv[i], "-h") == 0) {
			print_help();
			return 0;
		} else if (strcmp(argv[i], "-i") == 0) {
			if ((i + 2) >= argc) {
				fprintf(stderr,
					"must have host and number to increment\n");
				return -1;
			}
			increment_mode = 1;
			snprintf(host, sizeof(host), "%s", argv[i + 1]);
			increment = strtoll(argv[i + 2], &endptr, 10);
			if (*endptr != 0 || increment < 0) {
				fprintf(stderr, "illegal increment: %s\n",
					argv[i + 2]);
				return -1;
			}
			i += 3;
		} else if (argv[i][0] == '-') {
			fprintf(stderr,
				"unknown switch: %s. Use -h for help.\n",
				argv[i]);
			return -1;
		} else {
			break;
		}
	}
	if (i == argc) {
		if (getenv("HOME") != NULL) {
			snprintf(file, sizeof(file), "%s/.stt/default",
				 getenv("HOME"));
		} else {
			fprintf(stderr,
				"No HOME environment variable to lookup default host file.\n");
			return -1;
		}
	} else if ((i + 1) == argc) {
		snprintf(file, sizeof(file), "%s", argv[i]);
	} else {
		fprintf(stderr, "Please give only 1 host file argument\n");
		return -1;
	}

	if (increment_mode == 0) {
		long long bytes;
		if ((bytes = get_account(host, sizeof(host), file)) < 0) {
			fprintf(stderr,
				"no hosts found from host file %s\n",
				file);
			return -1;
		}
		if (print_bytes == 0) {
			printf("%s\n", host);
		} else {
			printf("%s %lld\n", host, bytes);
		}
	} else {
		account_increment(host, file, increment);
	}
	return 0;
}
